(*                                                                          *)
(* (c) 2004, Anastasia Gornostaeva. <ermine@ermine.pp.ru>                   *)
(*                                                                          *)

exception HttpClientError of string

let chunked_rex = Str.regexp "\\([a-f0-9]+\\)"
let get_chunked_size in_http =
   let line = input_line in_http in
      if Str.string_match chunked_rex line 0 then
	 let hex = Str.matched_group 1 line in
	    Hex.hex_to_int hex
      else
	 raise (HttpClientError ("Bad chunked size " ^ line))

let read_chunked in_http =
   let rec cycle acc =
      let chunked_size = get_chunked_size in_http in
	 if chunked_size = 0 then acc
	 else
	    let buffer = String.create chunked_size in
	       really_input in_http buffer 0 chunked_size;
	       let skip = input_line in_http in
		  cycle (acc ^ buffer)
   in
      cycle ""

let get_header headers key =
   let (_, value) = List.find (function (key1, value) ->
				  key1 = key
			      ) headers in
      value

let header = Str.regexp "\\([a-zA-Z-]+\\): \\([^\r\n]+\\)"

let read_headers in_http =
   let rec read acc =
      let line = input_line in_http in
	 if line = "\r" then 
	    acc
	 else
	    if Str.string_match header line 0 then
	       let key = Str.matched_group 1 line in
	       let value = Str.matched_group 2 line in
		  read ((key, value) :: acc)
	    else
	       failwith ("Error parsing header: " ^ line)
   in
      read []

let connect server port =
   let inet_addr =
      try Unix.inet_addr_of_string server with Failure("inet_addr_of_string") ->
         (Unix.gethostbyname server).Unix.h_addr_list.(0) in
   let sock_addr = Unix.ADDR_INET (inet_addr, port) in
      try
	 Unix.open_connection sock_addr
      with Unix.Unix_error (errno, "connect", _) ->
	 raise (HttpClientError (Printf.sprintf "Cannot connect to %s:%d: %s"
				    server port
				    (Unix.error_message errno)))

type http_method = | Post of string * string | Get of string

let request server http_method headers =
   let in_http, out_http = connect server 80 in
      (match http_method with
	 | Post (path, data) ->
	      let length = string_of_int (String.length data) in
		 output_string out_http ("POST " ^ path ^ " HTTP/1.1\r\n");
		 output_string out_http ("HOST: " ^ server ^ "\r\n");
		 output_string out_http ("Content-Length: " ^ length ^ "\r\n");
		 List.iter (function header ->
			       output_string out_http header;
			       output_string out_http "\r\n"
			   ) headers;
		 output_string out_http "\r\n";
		 output_string out_http data;
		 flush out_http;
	 | Get path ->
	      output_string out_http ("GET " ^ path ^ " HTTP/1.1\r\n");
	      output_string out_http ("HOST: " ^ server ^ "\r\n");
	      List.iter (function header ->
			    output_string out_http header;
			    output_string out_http "\r\n"
			) headers;
	      output_string out_http "\r\n";
	      flush out_http);
      let status = input_line in_http in
	 if status = "HTTP/1.1 200 OK\r" then
	    let headers = read_headers in_http in
	       try
		  let chunked = get_header headers "Transfer-Encoding" in
		     if chunked = "chunked" then
			let content = read_chunked in_http in
			   close_in in_http;
			   content
		     else
			raise (HttpClientError 
				  ("Unknown transfer-encoding: " ^ chunked))
	       with Not_found ->
		  let length = get_header headers "Content-Length" in
		  let buffer = String.create (int_of_string length) in
		     really_input in_http buffer 0 (int_of_string length);
		     buffer
	 else
	    raise (HttpClientError ("Not good response status: " ^ status))

