(** timer library. This library uses Sys module and unix, thread library. *)

open Sys
open Unix
open Event

(** default tick unit in milli seconds.
    This is used in tick server.
*)
let tick_unit = ref 100.0

(** callback id.
*)
type id = unit ref

(** callback item.
*)
type callback = {
  f: unit -> unit;
  time: float;
  id: id
}

(** protocols between user API and timer server.
*)
type protocol =
  | Register of
      (unit -> unit) *		(* callback function *)
      float *			(* callback start time in milli-seconds *)
      float *			(* callback interval in milli-seconds *)
      id channel		(* channel to return callback id *)
  | Unregister of id
  | Tick

let send_sync ch v = sync (send ch v)
let recv_sync ch = sync (receive ch)

(** timer server.
*)
let the_server = ref None

(** API to register callback.
{e register f start interval} registers callback fucntion
{e f} which will be called after {e start} milli seconds at first
and every {e interval} milli seconds after the first calling.
If {e interval} is zero, the callback function is called only once.
@param f callback function of type of unit -> unit
@param start next expiration time delay in milli-seconds
@param interval interval time in milli-seconds
@return id the registered callback id.
*)
let register f start interval =
  match !the_server with
  | None -> failwith "not initialized yet"
  | Some (_, ch) ->
      let reply_ch = new_channel () in
      send_sync ch (Register (f, start, interval, reply_ch));
      recv_sync reply_ch

(** sleep for the given time.
@param msec sleep time in milli-seconds
*)
let msleep msec =
  let ch = new_channel () in
  let f () = send_sync ch () in
  register f msec 0.0;
  recv_sync ch

(** unregister a callback with the given id.
*)
let unregister id =
  match !the_server with
  | None -> failwith "not initialized yet"
  | Some (_, ch) ->
      send_sync ch (Unregister id)

(** send tick to timer server.
*)
let send_tick signal =
  match !the_server with
  | None -> failwith "not initialized yet"
  | Some (_, ch) -> send_sync ch Tick

(** initialize tick server.
*)
let init_tick_server ?(tck_unit = !tick_unit) () =
  tick_unit := tck_unit;
  let _ = set_signal sigalrm (Signal_handle send_tick) in
  let _ = setitimer ITIMER_REAL {it_interval = tck_unit /. 1000.; it_value = tck_unit /. 1000.} in
  ()

(** reset tick server.
*)
let reset_tick_server () =
  let _ = set_signal sigalrm Signal_ignore in
  let _ = setitimer ITIMER_REAL {it_interval = 0.; it_value = 0.} in
  ()

(** set timer server and initialize tick server.
*)
let set_server ?tck_unit server =
  the_server := Some server;
  init_tick_server ?tck_unit ()

(** insert a callback structure to timer waiting list.
*)
let insert waiting_list cb =
  let rec loop cbs =
    match cbs with
    | [] -> [cb]
    | x::xs ->
        if x.time < cb.time then x :: loop xs
        else cb :: cbs
  in
  waiting_list := loop !waiting_list

(** remove an item with the given id from the timer waiting list.
*)
let remove waiting_list id =
  let rec loop cbs =
    match cbs with
    | [] -> (* Not found *)
	[]
    | x::xs ->
        if x.id == id	(* Physical Equality Test *)
        then xs
        else x :: loop xs
  in
  waiting_list := loop !waiting_list

(** get expired item from timer waiting list.
@param waiting_list timer waiting list
*)
let get_expired waiting_list =
  let current = gettimeofday () in
  let threshold = current +. !tick_unit /. 1000.0 /. 10.0 in
  let rec loop cbs expired =
    match cbs with
    | [] -> (List.rev expired, [])
    | x::xs ->
        if x.time <= threshold
        then loop xs (x::expired)
	else (List.rev expired, cbs)
  in
  let (expired, new_waiting) = loop !waiting_list [] in
  waiting_list := new_waiting;
  expired

(** run the function forever.
@param f a function
@param arg argument to the function
*)
let forever f arg =
  let rec loop () =
    f arg;
    loop ();
  in
  loop ()

(** make a callback structure.
*)
let make_cb f start interval waiting_list =
  let id = ref () in
  let exp_time = gettimeofday () +. start /. 1000.0 in
  if interval <= 0.0
  then
    {f = f; time = exp_time; id = id }
  else
    let rec interval_f () =
      let next_time = gettimeofday () +. interval /. 1000.0 in
      f ();
      let cb = {f = interval_f; time = next_time; id = id } in
      insert waiting_list cb
    in
    {f = interval_f; time = exp_time; id = id }

(** main function of the timer server.
@param ch communication channel to timer server
*)
let server_main ch =
  let waiting_list = ref [] in
  let loop () =
    let msg = recv_sync ch in
    begin match msg with
    | Register (f, start, interval, reply_ch) ->
        let cb = make_cb f start interval waiting_list in
        insert waiting_list cb;
        ignore (send_sync reply_ch cb.id)
    | Unregister id ->
        remove waiting_list id
    | Tick ->
	let expired = get_expired waiting_list in
        let invoke cb = cb.f () in
	List.iter invoke expired
    end;
  in
  forever loop ()

(** initializer timer library.
    This should be called before calling any timer function.
    @param tck_unit tick unit of tick timer. (default value is 100 milli seconds)
*)
let init ?tck_unit () =
  let ch = new_channel () in
  let tid = Thread.create server_main ch in
  let server = (tid, ch) in
  set_server ?tck_unit server

(** reset timer library.
*)
let reset () =
  reset_tick_server ();
  match !the_server with
  | None -> ()
  | Some (tid, _) -> Thread.kill tid
